#!/usr/bin/env python

from __future__ import print_function
import os
import sys
import urllib
import urllib2
import urlparse
import signal
import argparse
import logging

import socket
# this controls buffering in the urllib2 interface, which we want to
# eliminate so that stream data will be processed immediately.
# https://stackoverflow.com/questions/1598331/how-to-read-continous-http-streaming-data-in-python
socket._fileobject.default_bufsize = 0

from . import const
from . import command

GUARDCTRL_HOST = os.getenv('GUARDCTRL_HOST')
if not GUARDCTRL_HOST:
    IFO = os.getenv('IFO','').upper()
    if IFO == '':
        sys.exit("Must specify either IFO or GUARDCTRL_HOST environment variable.")
    elif IFO == 'H1':
        GUARDCTRL_HOST = 'h1guardian0'
    elif IFO == 'L1':
        GUARDCTRL_HOST = 'l1script0'
    else:
        GUARDCTRL_HOST = IFO.lower()+'guardian0'

##################################################

class GuardCtrlClientError(Exception):
    pass

class GuardCtrlClient(object):
    def __init__(self, host=GUARDCTRL_HOST, port=const.GUARDCTRL_PORT):
        self.host = host
        self.port = port

    def format_url(self, cmd, **kwargs):
        return urlparse.ParseResult(
            scheme='http',
            netloc='{}:{}'.format(self.host, self.port),
            path='/{}/'.format(cmd),
            params='',
            query=urllib.urlencode(kwargs, doseq=True),
            fragment='',
            ).geturl()

    def _get(self, cmd, **kwargs):
        """Generator of lines in command query response

        """
        url = self.format_url(cmd, **kwargs)
        logging.debug("query: {}".format(url))
        request = urllib2.Request(url)

        try:
            response = urllib2.urlopen(request)
        # NOTE: this has to be caught before URLError, since apparently this inherits from URLError?
        except urllib2.HTTPError as e:
            raise GuardCtrlClientError("Request error {}: {}".format(e.code, e.reason))
        except urllib2.URLError as e:
            raise
            base = urlparse.urlunsplit(urlparse.urlsplit(url)[0:2]+('','',''))
            raise GuardCtrlClientError("Could not contact guardctrl server: {}".format(base))

        line = None
        for line in response:
            yield(line.strip())

        if line and line.strip() == response.info().getheader(const.SERVER_ERROR_HEADER):
            raise GuardCtrlClientError("Server returned error during processing.  Please inform admin.")

    def cmd(self, args):
        """Execute command from argparse Namespace object

        Yields lines of the response from the specified query.

        """
        query = {k:v for k,v in vars(args).iteritems() if v and k != 'cmd'}
        logging.debug("query: {}".format(query))
        return self._get(args.cmd, **query)

    def node_list(self, **kwargs):
        """Return list of node names

        """
        return [line for line in self._get('list', **kwargs)]

    def log(self, nodes, **kwargs):
        """Generator of log lines matching search parameters

        `nodes` is a list of node names of glob patterns

        """
        kwargs['nodes'] = nodes
        for line in self._get('log', **kwargs):
            yield line

##################################################

def xlaunch(cmd, xargs=[]):
    """Launch command in background xterm

    First argument is command to be run in list form.  'xargs' is list
    of additional arguments to pass to xterm.

    """
    title = ' '.join(cmd)
    xcmd = ['xterm', '-si', '-sk',
            '-geometry', '160x80',
            '-bg', 'DodgerBlue4',
            '-title', title,
            ] + xargs + ['-e'] + cmd
    logging.debug(xcmd)
    os.spawnvp(os.P_NOWAIT, 'xterm', xcmd)

##################################################

if __name__ == '__main__':

    prog = 'guardctrl'
    description = """Guardian process control client

Add '-h' after subcommand for subcommand help.
"""
    parser = argparse.ArgumentParser(prog=prog,
                                     description=description,
                                     formatter_class=argparse.RawDescriptionHelpFormatter,
                                     )
    parser.add_argument('-x', action='store_true',
                        help="execute command in X")
    parser.add_argument('--debug-query', action='store_true', help=argparse.SUPPRESS)

    subparser = parser.add_subparsers(
        metavar='COMMAND',
        dest='cmd',
        )

    for cmd, mod in command.command_modules.iteritems():
        sp = subparser.add_parser(cmd,
                                  help=mod.SUMMARY,
                                  formatter_class=argparse.RawDescriptionHelpFormatter,
                                  )
        mod.gen_argparser(sp)

    args = parser.parse_args()
    cmd = args.cmd

    logging.debug('args: %s' % (args))

    ##########
    if args.x:
        xlaunch([prog] + [a for a in sys.argv[1:] if a != '-x'])
        sys.exit()
    del args.x
    ##########

    signal.signal(signal.SIGINT, signal.SIG_DFL)
    signal.signal(signal.SIGPIPE, signal.SIG_DFL)

    try:
        for line in GuardCtrlClient().cmd(args):
            print(line)
    except GuardCtrlClientError as e:
        sys.exit(e)
