import os
import re
import logging
import argparse
import datetime
import subprocess
from datetime import timedelta
from gpstime import gpstime
from dateutil.tz import tzutc, tzlocal

from .. import const
from .. import util

##################################################

TIME_RE = re.compile('(\d{4}-?\d{2}-?\d{2}[T_]\d{2}:\d{2}:\d{2}\S*) ?(.*)')

##################################################

def find_log_files(logdir):
    """generator of all svlogd log files in directory"""
    # for python3:
    # with subprocess.Popen(['/usr/bin/tai64nlocal'],
    #                       stdin=subprocess.PIPE,
    #                       stdout=subprocess.PIPE,
    #                       #universal_newlines=True,
    #                       ) as p:
    # FIXME: this is obviously non-portable
    p = subprocess.Popen(['/usr/bin/tai64nlocal'],
                         stdin=subprocess.PIPE,
                         stdout=subprocess.PIPE,
                         )
    if True:
        for f in os.listdir(logdir):
            if f[0] != '@':
                continue
            i = f.split('.')[0]
            p.stdin.write(i+'\n')
            o = p.stdout.readline().strip().split('.')[0]
            dt = gpstime.strptime(o, '%Y-%m-%d %H:%M:%S')
            # tai64n log file names are local time
            dt = dt.replace(tzinfo=tzlocal())
            yield (f, dt)
        p.terminate()
        p.wait()
        if os.path.exists(os.path.join(logdir, 'current')):
            # include 'current' at the end
            yield ('current', gpstime.now(tz=tzlocal()))

def filter_files(logdir, after=None, before=None):
    """generator of svlogd log files from directory covering interval (after,before)"""
    last = gpstime.fromtimestamp(0, tz=tzutc())
    for f, dt in sorted(find_log_files(logdir), key=lambda k: k[1]):
        bb = True
        ab = True
        if after and after > dt:
            bb = False
        if before and before < last:
            ab = False
        last = dt
        if bb and ab:
            yield (f, dt)

def filter_files_multidir(logdirs, after=None, before=None):
    """generator of all svlogd files in a list of directories, filtered by time interval"""
    for logdir in logdirs:
        for f, dt in filter_files(logdir, after, before):
            yield (os.path.join(logdir, f), dt)

def parse_log_line(line):
    """parse log line

    Take a log line string and return a tuple of (datetime, msg)
    extracted from the line.

    """
    line = line.strip()
    match = TIME_RE.match(line)
    if not match:
        return None, line
    ds = match.group(1)
    msg = match.group(2)
    # dt = gpstime.strptime(ds.split('.')[0], '%Y-%m-%dT%H:%M:%S')
    try:
        dt = gpstime.strptime(ds, '%Y-%m-%dT%H:%M:%S.%f')
    except ValueError:
        try:
            dt = gpstime.strptime(ds, '%Y-%m-%dT%H:%M:%S.%fZ')
        except ValueError:
            try:
                dt = gpstime.strptime(ds, '%Y%m%d_%H:%M:%S.%f')
            except:
                return None, msg
    # guardian log times are UTC
    dt = dt.replace(tzinfo=tzutc())
    return dt, msg

def dump_lines_multifile(fs):
    """generator of log lines from specified files"""
    for f in fs:
        logging.debug("dumping file: %s" % f)
        with util.opportunistically_decompress(f) as fd:
            for line in fd:
                yield line.strip()

def filter_lines(lines, after=None, before=None):
    """filter lines within specified range"""
    for line in lines:
        dt, msg = parse_log_line(line)
        if dt:
            if util.in_interval(dt, after, before):
                yield (dt, msg)
        else:
            yield (dt, msg)

def filter_lines_multifile(fs, after=None, before=None):
    """generator of log lines within specified range from specified files"""
    for f in fs:
        logging.debug("filtering file: %s" % f)
        with util.opportunistically_decompress(f) as fd:
            for line in filter_lines(fd, after, before):
                yield line

##################################################

def gen_node_log_dirs(nodes):
    """generator of node log directory paths"""
    for node in nodes:
        yield os.path.join(const.GUARD_NODE_LOG_DIR, node)

def find_node_log_files(nodes, after=None, before=None):
    """find all log files for the specified nodes within time range"""
    nodedirs = list(gen_node_log_dirs(nodes))
    if after is True:
        return [os.path.join(d, 'current') for d in nodedirs]
    else:
        return [p for p,dt in sorted(filter_files_multidir(nodedirs, after, before), key=lambda k: k[1])]

def sort_log_lines(lines, sort='time'):
    """sort log lines by date"""
    if sort == 'time':
        return sorted(lines, key=lambda k: k[0])
    else:
        return lines

def format_log_line(dt, msg, time_format='utc'):
    """format output of log (datetime, message) tuples"""
    if dt:
        if time_format == 'utc':
            dts = dt.strftime('%Y-%m-%d_%H:%M:%S.%fZ')
        elif time_format == 'local':
            dts = dt.astimezone(tzlocal()).strftime('%Y-%m-%d_%H:%M:%S.%f')
        elif time_format == 'gps':
            dts = '{:.3f}'.format(dt.gps())
        else:
            dts = str(dt)
    else:
        dts = '??????????'
    return '{} {}'.format(dts, msg)

##################################################

def node_log_follow_tail(nodes, time_format='raw'):
    """follow node log files

    Outputs the last 100 lines of each node log, then follow the logs
    as they appear.

    """
    cfiles = find_node_log_files(nodes, after=True)
    # retrieve last 100 lines from all the log files
    # HACK: this shouldn't need to be a subprocess call
    cmd = ['tail', '-q', '--lines=100'] + cfiles
    #logging.debug(' '.join(cmd))
    p0 = subprocess.Popen(cmd, stdout=subprocess.PIPE)
    try:
        for line in sorted(iter(p0.stdout.readline, '')):
            if time_format == 'raw':
                yield line.strip()
            else:
                dt, msg = parse_log_line(line)
                yield format_log_line(dt, msg, time_format)
    except SystemExit:
        pass
    finally:
        p0.wait()
    # now actually follow the files
    # FIXME: how to do this pythonically
    cmd = ['tail', '-F', '-q', '--lines=0'] + cfiles
    #logging.debug(' '.join(cmd))
    p1 = subprocess.Popen(cmd, stdout=subprocess.PIPE)
    try:
        for line in iter(p1.stdout.readline, ''):
            if time_format == 'raw':
                yield line.strip()
            else:
                dt, msg = parse_log_line(line)
                yield format_log_line(dt, msg, time_format)
    except SystemExit:
        pass
    finally:
        p1.wait()

def node_log_search(nodes, after=None, before=None, time_format='utc'):
    """generator of node log lines in some time range

    """
    paths = find_node_log_files(nodes, after, before)

    # after True is used to indicate a search for the current log
    # file, in which case we don't filter lines
    if after is True:
        after = None

    # opportunistically sort the lines only if more then one node is
    # specified
    if len(nodes) == 1:
        sort = None
    else:
        sort = 'time'

    # output lines
    if time_format == 'raw':
        for line in dump_lines_multifile(paths):
            yield line
    else:
        it = filter_lines_multifile(paths, after, before)
        for dt, msg in sort_log_lines(it, sort):
            yield format_log_line(dt, msg, time_format)

##################################################

SUMMARY = "view/search node logs"

def gen_argparser(parser=None, **kwargs):
    if not parser:
        parser = argparse.ArgumentParser(**kwargs)

    parser.formatter_class = argparse.RawDescriptionHelpFormatter
    parser.description = SUMMARY + """

With no search filters, the logs of the specified nodes will be
followed.  If search filters are specified, logs corresponding to the
specified search will be output without following.  Log timestamps are
in UTC by default.

"""
    parser.epilog = """
Date/time strings for filters are handled by a date/time string parser
and therefore may be specified in various formats.  If --after and
--before are not both specified, --duration will be used.

Examples:

  Show all ISC_LOCK logs from Sept. 14, 2015:

  $ {prog} --after 2015-09-14 --before 2015-09-15 ISC_LOCK

  Use wildcard to show all SUS logs from today:

  $ {prog} --today SUS_*

  Quote compound date strings, and specify 'UTC' for times in UTC:

  $ {prog} -a "2016-08-20 06:23:22 UTC" -b "2016-08-20 06:23:52 UTC" ISC_LOCK

  GPS times are accepted as well:

  $ {prog} -c 1155709419 ISC_DRMI

""".format(prog=parser.prog)

    parser.add_argument('nodes', metavar='NODE', nargs='*',
                        help="node name (wildcard/globbing supported)")

    tgroup = parser.add_mutually_exclusive_group()
    tgroup.add_argument('-a', '--after', metavar='DATETIME',
                        help="filter for logs after specified date/time")
    tgroup.add_argument('-c', '--center', metavar='DATETIME',
                        help="filter for logs centered around specified time")
    tgroup.add_argument('-o', '--hours', metavar='H', type=int, dest='hours',
                        help="filter for logs from the last H hours")
    tgroup.add_argument('-t', '--today', action='store_const', dest='hours', const=24,
                        help="filter for logs from the last 24 hours")
    tgroup.add_argument('-n', '--current', action='store_true',
                        help="filter for logs from current log file (fast)")

    parser.add_argument('-b', '--before', metavar='DATETIME',
                        help="filter for logs before specified date/time (requires --after)")
    parser.add_argument('-d', '--duration', metavar='S', type=float, default=60,
                        help="duration after/before/around-center, in seconds [60]")

    ogroup = parser.add_mutually_exclusive_group()
    ogroup.add_argument('-u', '--utc', action='store_const', dest='output', const='utc',
                        help="output log lines with UTC timestamps (default)")
    ogroup.add_argument('-l', '--local', action='store_const', dest='output', const='local',
                        help="output log lines with local timestamps")
    ogroup.add_argument('-g', '--gps', action='store_const', dest='output', const='gps',
                        help="output log lines with GPS timestamps")
    ogroup.add_argument('-r', '--raw', action='store_const', dest='output', const='raw',
                        help="output contents of matched files without line parsing/sorting (fastest)")
    ogroup.add_argument('-p', '--paths', action='store_const', dest='output', const='paths',
                        help="output just log file paths")

    #parser.add_argument('--help', action='help', help=argparse.SUPPRESS)

    return parser

def validate_args(args):
    if 'nodes' not in args:
        raise util.GuardCtrlError("must specify NODE")

    # check argument
    def carg(arg):
        return hasattr(args, arg) and getattr(args, arg)

    # parse times
    def _parse_time_arg(name):
        if not carg(name):
            return None
        string = getattr(args,name)
        try:
            t = gpstime.parse(string)
        except ValueError:
            raise util.GuardCtrlError("could not parse time string: '%s'" % string)
        return t

    ddelta = datetime.timedelta(seconds=float(args.duration))

    # parse time args
    if carg('current'):
        args.after = True
        args.before = None
    elif carg('hours'):
        args.after = gpstime.now(tz=tzlocal()) -  datetime.timedelta(seconds=3600*float(args.hours))
        args.before = None
    elif carg('after'):
        args.after = _parse_time_arg('after')
        if carg('before'):
            args.before = _parse_time_arg('before')
            if args.before < args.after:
                raise util.GuardCtrlError("before time is earlier than after time")
        elif carg('duration'):
            args.before = args.after + ddelta
    elif carg('center'):
        dur = ddelta/2
        args.after = _parse_time_arg('center') - dur
        args.before = _parse_time_arg('center') + dur
    elif carg('before'):
        args.before = _parse_time_arg('before')
        args.after = args.before - ddelta
    else:
        args.after = None
        args.before = None

    # parse output args
    if carg('after') or (carg('output') and args.output in ['paths', 'raw']):
        args.follow = False
        if not carg('output'):
            args.output = 'utc'
    else:
        args.follow = True
        if not carg('output'):
            args.output = 'raw'

    logging.debug("search after : %s" % args.after)
    logging.debug("search before: %s" % args.before)

    return args

def execute_args(args):
    if args.follow:
        return node_log_follow_tail(args.nodes, time_format=args.output)
    else:
        if args.output == 'paths':
            return find_node_log_files(args.nodes, args.after, args.before)
        else:
            return node_log_search(args.nodes, args.after, args.before, time_format=args.output)
